#include "parti.h"
#include "vararg.h"

/* variable-length arguments 
 *
 * new_mulscatter(sched, func, size, narray, data, target, ...)
 * SCHED *sched;		schedule
 * void  (*func)();		function
 * int   size;			size of each element
 * int   narray;		number of arrays
 * char  *data,*target;		arrays of data arrays
 */
void 
PARTI_mulscatter VA_DCL (void *, args)
{
  int   i, j;
  int   size;
  int   narray;
  char  **data, **target;
  void  (*func)();
  SCHED *sched;
  va_list ap;
  void  new_multi_scatter();

  /* retrieve the parameters from the list */
  sched  = (SCHED *) VA_START (ap, args, void *);
  func   = (void *) va_arg (ap, int);
  size   = (int) va_arg (ap, int);
  narray = (int) va_arg (ap, int);
  data   = (char **) malloc (sizeof(char *)*narray);
  target = (char **) malloc (sizeof(char *)*narray);
  for (i = 0; i < narray; i++) {
    data[i]   = (char *) va_arg (ap, char *);
    target[i] = (char *) va_arg (ap, char *);
  }

  /* call the multi_scatter function */
  new_multi_scatter(sched, func, size, narray, data, target);

  /* free the allocated memory */
  free (target);
  free (data);
}

/*
 *  wrapper for the mulscatter function
 */

extern SCHED *get_sched_ptr();

void 
parti_mulscatter_ VA_DCL (void *, args)
{
  int   i, j;
  int   *size;
  int   *func;
  int   *narray;
  int   *sched;
  char  **data, **target;
  va_list ap;
  void  new_multi_gather();

  /* retrieve the parameters from the list */
  sched  = (int *) VA_START (ap, args, int *);
  func   = (int *) va_arg (ap, int *);
  size   = (int *) va_arg (ap, int *);
  narray = (int *) va_arg (ap, int *);
  data   = (char **) malloc (sizeof(char *)*(*narray));
  target = (char **) malloc (sizeof(char *)*(*narray));
  for (i = 0; i < *narray; i++) {
    data[i]   = (char *) va_arg (ap, char *);
    target[i] = (char *) va_arg (ap, char *);
  }

  /* call the multi_scatter function */
  new_multi_scatter((SCHED *)get_sched_ptr(*sched), *func, *size,
							*narray, data, target);

  /* free the allocated memory */
  free (target);
  free (data);
}

void
new_multi_scatter(sched, func, size, narray, data, target)
SCHED *sched;			/* schedule */
void  (*func)();		/* function */
int   size;			/* size of each element */
int   narray;			/* number of arrays */
char  **data,**target;		/* arrays of data arrays */
{
  int  count,i,j,k,n,m;
  char **blist,*buffer;
  char *ptr;
  int  *ibuf, *idata;
  double *dbuf, *ddata;
  double start, end;

#ifdef DEBUG
  trace_enter("new_mulscatter");
#endif

  blist = wbuffer->blist;
  buffer = (char *) malloc(size*narray*sched->ndata);
  for(i=0, count = 0; i<PARTI_numnodes; i++){
    blist[i] = &buffer[count];
    count += size * sched->sr->fetch_n[i] * narray;
  }

/*
  for(i=0, ptr=buffer; i<sched->ndata; i++){
    for (j = 0; j < narray; j++, ptr+=size)
      bytecopy(ptr, &data[j][size*sched->perm[i]], size);
  }
*/
  switch (size) {
  case sizeof(int):
    ibuf = (int *) buffer;
    for (i = 0, count = 0, m = 0; i < PARTI_numnodes; i++) {
      for (j = 0; j < narray; j++) {
        idata = (int *) data[j];
	if (sched->type == NC) {
          for(k = 0, n = count; k < sched->sr->fetch_n[i]; k++, n++, m++)
             ibuf[m] = idata[sched->local_refs[sched->perm[n]]];
	}
	else {
          for(k = 0, n = count; k < sched->sr->fetch_n[i]; k++, n++, m++)
             ibuf[m] = idata[sched->perm[n]];
	}
      }
      count += sched->sr->fetch_n[i];
    }
    break;
  case sizeof(double):
    dbuf = (double *) buffer;
    for (i = 0, count = 0, m = 0; i < PARTI_numnodes; i++) {
      for (j = 0; j < narray; j++) {
        ddata = (double *) data[j];
	if (sched->type == NC) {
          for(k = 0, n = count; k < sched->sr->fetch_n[i]; k++, n++, m++)
            dbuf[m] = ddata[sched->local_refs[sched->perm[n]]];
	}
	else {
          for(k = 0, n = count; k < sched->sr->fetch_n[i]; k++, n++, m++)
            dbuf[m] = ddata[sched->perm[n]];
	}
      }
      count += sched->sr->fetch_n[i];
    }
    break;
  case sizeof(char):
    for (i = 0, count = 0, m = 0; i < PARTI_numnodes; i++) {
      for (j = 0; j < narray; j++) {
	if (sched->type == NC) {
          for(k = 0, n = count; k < sched->sr->fetch_n[i]; k++, n++, m++)
            buffer[m] = data[j][sched->local_refs[sched->perm[n]]];
	}
	else {
          for(k = 0, n = count; k < sched->sr->fetch_n[i]; k++, n++, m++)
            buffer[m] = data[j][sched->perm[n]];
	}
      }
      count += sched->sr->fetch_n[i];
    }
    break;
  default:
    for (i = 0, ptr = buffer, count = 0; i < PARTI_numnodes; i++) {
      for (j = 0; j < narray; j++) {
	if (sched->type == NC) {
          for(k = 0, n = count; k < sched->sr->fetch_n[i]; k++, n++, ptr+=size)
  	    bytecopy(ptr, &data[j][size*sched->local_refs[sched->perm[n]]], size);
	}
	else {
          for(k = 0, n = count; k < sched->sr->fetch_n[i]; k++, n++, ptr+=size)
  	    bytecopy(ptr, &data[j][size*sched->perm[n]], size);
	}
      }
      count += sched->sr->fetch_n[i];
    }
    break;
  }

  new_multi_execute_scatter(sched->sr, func, size, narray, blist, target);

  free(buffer);

#ifdef DEBUG
  trace_exit();
#endif
}
